# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.


import logging
import re
from collections import defaultdict
from typing import List, Optional, Union

import h5py
import numpy as np

from threedigrid.admin.breaches.models import Breaches
from threedigrid.admin.breaches.timeseries_mixin import (
    BreachesAggregateResultsMixin,
    BreachesResultsMixin,
    get_breaches_customized_result_mixin,
)
from threedigrid.admin.constants import DEFAULT_CHUNK_TIMESERIES
from threedigrid.admin.gridadmin import GridH5Admin
from threedigrid.admin.h5py_datasource import H5pyResultGroup
from threedigrid.admin.h5py_swmr import H5SwmrFile
from threedigrid.admin.lines.models import Lines
from threedigrid.admin.lines.timeseries_mixin import (
    get_lines_customized_result_mixin,
    LinesAggregateResultsMixin,
    LinesResultsMixin,
)
from threedigrid.admin.nodes.models import Nodes
from threedigrid.admin.nodes.timeseries_mixin import (
    get_nodes_customized_results_mixin,
    get_nodes_customized_water_quality_results_mixin,
    get_substance_result_mixin,
    NodesAggregateResultsMixin,
    NodesResultsMixin,
)
from threedigrid.admin.pumps.models import Pumps
from threedigrid.admin.pumps.timeseries_mixin import (
    get_pumps_customized_result_mixin,
    PumpsAggregateResultsMixin,
    PumpsResultsMixin,
)
from threedigrid.admin.structure_controls.models import (
    StructureControl,
    StructureControlSourceTypes,
    StructureControlTypes,
)
from threedigrid.orm.models import Model

logger = logging.getLogger(__name__)

try:
    import asyncio_rpc  # noqa

    asyncio_rpc_support = True
except ImportError:
    asyncio_rpc_support = False


class GridH5ResultAdmin(GridH5Admin):
    """
    Admin interface for threedicore result queries.
    """

    def __init__(self, h5_file_path, netcdf_file_path, file_modus="r", swmr=False):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        self._field_model_dict = defaultdict(list)
        self._netcdf_file_path = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        self.result_datasource_class = H5pyResultGroup

        if h5_file_path.startswith("rpc://"):
            if not asyncio_rpc_support:
                raise Exception("Please reinstall this package with threedigrid[rpc]")

            from threedigrid.admin.rpc_datasource import H5RPCResultGroup, RPCFile

            self.netcdf_file = RPCFile(h5_file_path, file_modus)
            self.result_datasource_class = H5RPCResultGroup
        else:
            if swmr:
                self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
            else:
                self.netcdf_file = h5py.File(netcdf_file_path, file_modus)
            self.version_check()
        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)

    def set_timeseries_chunk_size(self, new_chunk_size):
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

    @property
    def timeseries_chunk_size(self):
        return self._timeseries_chunk_size.stop

    @property
    def time_units(self):
        return self.netcdf_file["time"].attrs.get("units")

    @property
    def lines(self):
        return Lines(
            self.result_datasource_class(self.h5py_file, "lines", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": LinesResultsMixin}),
        )

    @property
    def nodes(self):
        return Nodes(
            self.result_datasource_class(self.h5py_file, "nodes", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": NodesResultsMixin}),
        )

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info("Threedimodel has no breaches")
            return
        return Breaches(
            self.result_datasource_class(self.h5py_file, "breaches", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": BreachesResultsMixin}),
        )

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info("Threedimodel has no pumps")
            return
        return Pumps(
            self.result_datasource_class(self.h5py_file, "pumps", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": PumpsResultsMixin}),
        )

    def version_check(self):
        """
        compare versions of grid admin and grid results file.
        Issues a warning if they differ
        """
        return  # TODO implement a different check

        if self.threedicore_result_version != self.threedicore_version:
            logger.warning(
                "[!] threedicore version differ! \n"
                "Version result file has been created with: %s\n"
                "Version gridadmin file has been created with: %s",
                self.threedicore_result_version,
                self.threedicore_version,
            )

    @property
    def _field_model_map(self):
        """
        :return: a dict of {<field name>: [model name, ...]}
        """
        if self._field_model_dict:
            return self._field_model_dict

        model_names = set()
        for attr_name in dir(self):
            # skip private attrs
            if any([attr_name.startswith("__"), attr_name.startswith("_")]):
                continue
            try:
                attr = getattr(self, attr_name)
            except AttributeError:
                logger.warning(
                    "Attribute: '{}' does not " "exist in h5py_file.".format(attr_name)
                )
                continue
            if not issubclass(type(attr), Model):
                continue
            model_names.add(attr_name)

        for model_name in model_names:
            for x in getattr(self, model_name)._field_names:
                self._field_model_dict[x].append(model_name)
        return self._field_model_dict

    def get_model_instance_by_field_name(self, field_name):
        """
        :param field_name: name of a models field
        :return: instance of the model the field belongs to
        :raises IndexError if the field name is not unique across models
        """
        model_name = self._field_model_map.get(field_name)
        if not model_name or len(model_name) != 1:
            raise IndexError(
                "Ambiguous result. Field name {} yields {} model(s)".format(
                    field_name, len(model_name) if model_name else 0
                )
            )
        return getattr(self, model_name[0])

    @property
    def threedicore_result_version(self):
        """
        :return: version of the grid result file (if found), an empty
        string otherwise
        """
        try:
            return self.netcdf_file.attrs.get("threedicore_version")
        except AttributeError:
            logger.error(
                "Attribute threedicore_version could not be found in result file"
            )  # noqa
        return ""

    def close(self):
        super().close()
        self.netcdf_file.close()


class GridH5AggregateResultAdmin(GridH5ResultAdmin):
    """
    Admin interface for threedicore result queries.
    """

    def __init__(self, h5_file_path, netcdf_file_path, file_modus="r"):
        """

        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf result file (usually
            called subgrid_map.nc)
        :param file_modus: modus in which to open the files
        """
        super().__init__(h5_file_path, netcdf_file_path, file_modus)

    @property
    def lines(self):
        model_name = "lines"
        return Lines(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": LinesAggregateResultsMixin}),
        )

    @property
    def nodes(self):
        return Nodes(
            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": NodesAggregateResultsMixin}),
        )

    @property
    def breaches(self):
        if not self.has_breaches:
            logger.info("Threedimodel has no breaches")
            return
        model_name = "breaches"
        return Breaches(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": BreachesAggregateResultsMixin}),
        )

    @property
    def pumps(self):
        if not self.has_pumpstations:
            logger.info("Threedimodel has no pumps")
            return
        model_name = "pumps"
        return Pumps(
            H5pyResultGroup(self.h5py_file, model_name, self.netcdf_file),
            **dict(self._grid_kwargs, **{"mixin": PumpsAggregateResultsMixin}),
        )

    @property
    def time_units(self):
        logger.info("Time units are not defined globally for aggregated results")
        return None

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()


class GridH5StructureControl(GridH5Admin):
    """Interface for structure control netcdf

    This interface is different from the GridH5ResultAdmin and
    GridH5AggregateResultAdmin, as it resembles less the Nodes and Lines data compared
    to those interfaces. This interface should be viewed as a CSV file interface
    which can be used to extract the different control structures by control type and
    output data. It supplies interfaces to Nodes, Lines, Weirs, etc but does not
    provide the same timeseries filter capabilities as the GridH5ResultAdmin and
    GridH5AggregateResultAdmin.

        >>> gst = GridH5StructureControl(gridadmin_path, struct_control_actions_3di.nc)
        >>> gst.table_control
        >>> gst.timed_control
        >>> gst.memory_control
        >>> gst.table_control.id
        >>> struct_control = gst.table_control.group_by_id(ga.table_control.id[0])
        >>> struct_control.time
        >>> struct_control.action_value_1
    """

    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
    ):
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf structure control result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        self._netcdf_file_path: str = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)

    @property
    def table_control(self) -> "_GridH5NestedStructureControl":
        """Get the table control actions as ``_GridH5NestedStructureControl`` object"""
        return _GridH5NestedStructureControl(self, StructureControlTypes.table_control)

    @property
    def memory_control(self) -> "_GridH5NestedStructureControl":
        """Get the memory control actions as ``_GridH5NestedStructureControl`` object"""
        return _GridH5NestedStructureControl(self, StructureControlTypes.memory_control)

    @property
    def timed_control(self) -> "_GridH5NestedStructureControl":
        """Get the timed control actions as ``_GridH5NestedStructureControl`` object"""
        return _GridH5NestedStructureControl(self, StructureControlTypes.timed_control)

    def get_source_table(self, action_type, grid_id):
        """Get source_table and source_table_id based on action_type and grid_id"""
        if action_type == "set_pump_capacity":
            source_table = "v2_pumpstation"
            source_table_id = grid_id
        else:
            source_table = self.lines.content_type[grid_id].decode("utf-8")
            source_table_id = self.lines.content_pk[grid_id]

        return source_table, source_table_id

    def get_source_type(self, action_type):
        """Indicates whether the action is applied on a line, pump (or node) feature"""
        if action_type == "set_pump_capacity":
            return StructureControlSourceTypes.PUMPS
        elif action_type in [
            "set_discharge_coefficients",
            "set_crest_level",
            "set_gate_level",
        ]:
            return StructureControlSourceTypes.LINES
        else:
            raise NotImplementedError


class _GridH5NestedStructureControl:
    def __init__(
        self,
        structure_control: GridH5StructureControl,
        control_type: StructureControlTypes,
    ):
        """
        This class is not to be instantiated seperately.
        Calling the ``GridH5StructureControl`` properties ``table_control``, ``memory_control``,
        or ``timed_control`` returns a ``_GridH5NestedStructureControl`` instance.

        :param structure_control: GridH5StructureControl(GridH5ResultAdmin)
        :param control_type: str [table_control, memory_control, timed_control]
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the netcdf structure control result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        if control_type not in StructureControlTypes.__members__.values():
            raise ValueError(f"Unknown control type: {control_type}")

        self.struct_control = structure_control
        self.control_type: str = control_type.value

    @property
    def action_type(self) -> np.ndarray:
        """Get the action types"""
        # binary character arrays from Netcdf Fortran to numpy object arrays
        return np.array(
            [
                action_type.tobytes().decode("utf-8").strip()
                for action_type in self.struct_control.netcdf_file[
                    f"{self.control_type}_action_type"
                ][:]
            ],
            dtype=object,
        )

    @property
    def action_value_1(self) -> np.ndarray:
        """Get the action values"""
        return self.struct_control.netcdf_file[f"{self.control_type}_action_value_1"][:]

    @property
    def action_value_2(self) -> np.ndarray:
        """Get the second action values (negative discharge for ``action_type == 'set_discharge_coefficients'``)"""
        return self.struct_control.netcdf_file[f"{self.control_type}_action_value_2"][:]

    @property
    def grid_id(self) -> np.ndarray:
        """Get the grid ID's, i.e. the ID of the Node or Line upon which the structure control action acts"""
        return self.struct_control.netcdf_file[f"{self.control_type}_grid_id"][:]

    @property
    def id(self) -> np.ndarray:
        """Get the ID's of the structure control action"""
        # binary character arrays from Netcdf Fortran to numpy object arrays"
        return np.array(
            [
                id.tobytes().decode("utf-8").strip()
                for id in self.struct_control.netcdf_file[f"{self.control_type}_id"][:]
            ],
            dtype=object,
        )

    @property
    def is_active(self) -> np.ndarray:
        """Get the boolean values indicating if the structure control action is active"""
        return self.struct_control.netcdf_file[f"{self.control_type}_is_active"][:]

    @property
    def time(self) -> np.ndarray:
        """Get the times (in s since start of simulation) at which the structure controls acted"""
        return self.struct_control.netcdf_file[f"{self.control_type}_time"][:]

    def group_by_id(self, id: str) -> Optional[StructureControl]:
        """
        Get the structure control action with given ``id``.

        ID is unique. Get content_type and content_pk from gridadmin. All controls are
        on lines except set_pump_capacity
        """
        mask = np.isin(self.id, id)
        if np.all(~mask):
            return

        grid_id = self.grid_id[mask][0]
        action_type = self.action_type[mask][0]
        source_table, source_table_id = self.struct_control.get_source_table(
            action_type, grid_id
        )
        source_type = self.struct_control.get_source_type(action_type)

        return StructureControl(
            id=id,
            grid_id=grid_id,
            source_table=source_table,
            source_table_id=source_table_id,
            source_type=source_type,
            time=self.time[mask],
            action_type=action_type,
            action_value_1=self.action_value_1[mask],
            action_value_2=self.action_value_2[mask],
            is_active=self.is_active[mask],
        )

    def group_by_grid_id(self, value: int) -> List[StructureControl]:
        """
        Get all structure control actions with ``grid_id == value``
        """
        return self._group_by("grid_id", value)

    def group_by_action_type(self, value: str) -> List[StructureControl]:
        """
        Get all structure control actions with ``action_type == value``
        """
        return self._group_by("action_type", value)

    def group_by_is_active(self, value: int) -> List[StructureControl]:
        """
        Get all structure control actions with ``is_active == value``
        """
        return self._group_by("is_active", value)

    def group_by_time(self, min: float, max: float) -> List[StructureControl]:
        """Get all structure control actions where ``min <= time <= max``"""
        return self._group_by_in_between("time", min, max)

    def group_by_action_value_1(self, min: float, max: float) -> List[StructureControl]:
        """Get all structure control actions where ``min <= action_value_1 <= max``"""
        return self._group_by_in_between("action_value_1", min, max)

    def group_by_action_value_2(self, min: float, max: float) -> List[StructureControl]:
        """Get all structure control actions where ``min <= action_value_2 <= max``"""
        return self._group_by_in_between("action_value_2", min, max)

    def _group_by(
        self, type: str, value: Union[int, float, str]
    ) -> List[StructureControl]:
        """Group control action by a type, and sort them by unique id

        Args:
            type (str): is_active, action_type, id
            value (Union[int, float, str]): corresponding value for type

        Returns:
            List[StructureControl]: unique structures found from type, value combination
        """
        if not hasattr(self, type):
            return []

        values = getattr(self, type)
        mask = np.isin(values, value)
        ids = np.unique(self.id[mask])
        return [self.group_by_id(id) for id in ids]

    def _group_by_in_between(
        self, type: str, min: Union[int, float], max: Union[int, float]
    ) -> List[StructureControl]:
        """Find values between min and max

        Args:
            type (str): time, action_value_1, action_value_2
            value (Union[int, float, str]): corresponding value for type

        Returns:
            List[StructureControl]: unique structures found from type, value combination
        """
        if not hasattr(self, type):
            return []

        values = getattr(self, type)
        mask = (values >= min) & (values <= max)
        ids = np.unique(self.id[mask])
        return [self.group_by_id(id) for id in ids]


class GridH5WaterQualityResultAdmin(GridH5Admin):
    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
        substances: List[str] = [],
    ) -> None:
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the water quality result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        self._netcdf_file_path: str = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)

        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)

        # Set substances as attributes
        substances = set()
        for key in self.netcdf_file.keys():
            regex_match = re.search(r"substance\d+", key)
            if regex_match:
                substance = regex_match.group()
                if substance not in substances:
                    substances.add(substance)
                    self.__setattr__(
                        substance,
                        Nodes(
                            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
                            **dict(
                                self._grid_kwargs,
                                **{"mixin": get_substance_result_mixin(substance)},
                            ),
                        ),
                    )

                    # Set substance attributes on Nodes object
                    attrs_map = [("name", "substance_name"), ("units", "units")]
                    for attr, name in attrs_map:
                        value = self.netcdf_file[key].attrs.get(name)
                        if isinstance(value, bytes):
                            value = value.decode("utf-8")
                        elif isinstance(value, h5py.Empty):
                            value = ""
                        self.__getattribute__(substance).__setattr__(attr, value)
        self.substances = list(substances)

    def get_model_instance_by_field_name(self, field_name):
        """
        :param field_name: name of a models field
        :return: instance of the model the field belongs to
        :raises AttributeError if the model instance cannot be found
        """
        try:
            model_instance = self.__getattribute__(field_name)
            return model_instance
        except AttributeError:
            raise AttributeError(
                f"Model instance with field name {field_name} not found"
            )

    def set_timeseries_chunk_size(self, new_chunk_size: int) -> None:
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()


class CustomizedResultAdmin(GridH5Admin):
    """Interface for customized 3Di result files

    Customized 3Di result files are result files where users can specify nodes, lines,
    and pumps of interest, combined with specific output variables and intervals.
    This interface can be used to extract data from customized result files. It
    contains functions to extract data from nodes, lines, breaches, and pumps.
    Additionally, it can be used to extract data from the result area of interest.

        >>> cra = CustomizedResultAdmin(gridadmin_path, customized_results_3di.nc)
        >>> cra.nodes.id
        >>> cra.lines.id
        >>> cra.area1.nodes.subset("1D_ALL").id
        >>> cra.area1.lines.subset("2D_ALL").s1
        >>> cra.area1.breaches.id
        >>> cra.area2.pumps.id
    """

    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
    ) -> None:
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the water quality result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        super().__init__(h5_file_path, file_modus)

        self._netcdf_file_path: str = netcdf_file_path
        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)

        self.set_timeseries_chunk_size(DEFAULT_CHUNK_TIMESERIES.stop)
        self.netcdf_keys = self.netcdf_file.keys()

        self.areas = []
        for key in self.netcdf_keys:
            regex_match = re.search(r"Mesh(1|2)D(Node|Line|Pump)_id_area\d+", key)
            if regex_match:
                # Set area as attribute and create area result admin
                dataset_name = regex_match.group().split("_")[-1]
                area_name = self.netcdf_file[key].attrs.get("area_name", dataset_name)
                if not hasattr(self, dataset_name):
                    self.areas += [dataset_name]
                    self.__setattr__(
                        dataset_name,
                        _CustomizedAreaResultAdmin(self, dataset_name, area_name),
                    )

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()

    @property
    def nodes(self):
        """Build nodes interface if there are nodes present in the result file."""
        if not hasattr(self, "_nodes"):
            self._nodes = self._build_nodes_result_group("")
        return self._nodes

    @property
    def lines(self):
        """Build lines interface if there are lines present in the result file."""
        if not hasattr(self, "_lines"):
            self._lines = self._build_lines_result_group("")
        return self._lines

    @property
    def breaches(self):
        if not hasattr(self, "_breaches"):
            if not self.has_breaches:
                self._breaches = None
            else:
                self._breaches = self._build_breaches_result_group("")

        if self._breaches is None:
            logger.info("Threedimodel has no breaches")

        return self._breaches

    @property
    def pumps(self):
        if not hasattr(self, "_pumps"):
            if not self.has_pumpstations:
                self._pumps = None
            else:
                self._pumps = self._build_pumps_result_group("")

        if self._pumps is None:
            logger.info("Threedimodel has no pumps")

        return self._pumps

    @property
    def levees(self):
        return None

    @property
    def nodes_embedded(self):
        return None

    @property
    def cross_sections(self):
        return None

    @property
    def cells(self):
        return None

    def _build_nodes_result_group(self, area: str) -> Optional[Nodes]:
        size = self.result_group_size(f"nMesh2D_nodes{area}", f"nMesh1D_nodes{area}")
        if size == 0:
            return None

        return Nodes(
            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
            **dict(
                self._grid_kwargs,
                **{"mixin": get_nodes_customized_results_mixin(self.netcdf_keys, area)},
            ),
        )

    def _build_lines_result_group(self, area: str) -> Optional[Lines]:
        size = self.result_group_size(f"nMesh2D_lines{area}", f"nMesh1D_lines{area}")
        if size == 0:
            return None

        return Lines(
            H5pyResultGroup(self.h5py_file, "lines", self.netcdf_file),
            **dict(
                self._grid_kwargs,
                **{
                    "mixin": get_lines_customized_result_mixin(
                        self.netcdf_keys,
                        area,
                    )
                },
            ),
        )

    def _build_breaches_result_group(self, area: str) -> Breaches:
        return Breaches(
            H5pyResultGroup(self.h5py_file, "breaches", self.netcdf_file),
            **dict(
                self._grid_kwargs,
                **{
                    "mixin": get_breaches_customized_result_mixin(
                        self.netcdf_keys,
                        area,
                    )
                },
            ),
        )

    def _build_pumps_result_group(self, area: str) -> Pumps:
        return Pumps(
            H5pyResultGroup(self.h5py_file, "pumps", self.netcdf_file),
            **dict(
                self._grid_kwargs,
                **{
                    "mixin": get_pumps_customized_result_mixin(
                        self.netcdf_keys,
                        area,
                    ),
                },
            ),
        )

    def result_group_size(self, field_name_2d: str, field_name_1d) -> np.ndarray:
        two_d_size = self.netcdf_file.get(field_name_2d, np.array([])).size
        one_d_size = self.netcdf_file.get(field_name_1d, np.array([])).size
        return two_d_size + one_d_size

    def set_timeseries_chunk_size(self, new_chunk_size):
        """overwrite the default chunk size for timeseries queries"""
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})


class _CustomizedAreaResultAdmin:
    """Nested class for customized 3Di result files to interact with specific areas"""

    def __init__(
        self,
        customized_result_admin: CustomizedResultAdmin,
        dataset_name: str,
        area_name: str,
    ) -> None:
        self.cra = customized_result_admin
        self.dataset_name = dataset_name
        self.name = area_name
        self._timeseries_chunk_size = customized_result_admin._timeseries_chunk_size
        self._grid_kwargs = customized_result_admin._grid_kwargs

    @property
    def nodes(self):
        """Build nodes interface if there are nodes present in the result file."""
        if not hasattr(self, "_nodes"):
            self._nodes = self.cra._build_nodes_result_group(f"_{self.dataset_name}")
        return self._nodes

    @property
    def lines(self):
        """Build lines interface if there are lines present in the result file."""
        if not hasattr(self, "_lines"):
            self._lines = self.cra._build_lines_result_group(f"_{self.dataset_name}")
        return self._lines

    @property
    def breaches(self):
        if not hasattr(self, "_breaches"):
            self._breaches = self.cra._build_breaches_result_group(
                f"_{self.dataset_name}"
            )
        return self._breaches

    @property
    def pumps(self):
        if not hasattr(self, "_pumps"):
            self._pumps = self.cra._build_pumps_result_group(f"_{self.dataset_name}")
        return self._pumps


class CustomizedWaterQualityResultAdmin(GridH5Admin):
    """Interface for customized 3Di water quality result files

    Customized water quality 3Di result files are result files where users can specify
    nodes, lines, and pumps of interest. This interface can be used to extract data from
    customized water quality result files. It can be used to extract data from the
    result area of interest.

        >>> cwqa = CustomizedWaterQualityResultAdmin(gridadmin_path, customized_water_quality_results_3di.nc)
        >>> cwqa.substance1.concentration
        >>> cwqa.substance2.name
        >>> cwqa.area1.substance1.subset("1D_ALL").id
        >>> cwqa.area2.substance2.subset("2D_ALL").concentration
    """

    def __init__(
        self,
        h5_file_path: str,
        netcdf_file_path: str,
        file_modus: str = "r",
        swmr: bool = False,
    ) -> None:
        """
        :param h5_file_path: path to the hdf5 gridadmin file
        :param netcdf_file_path: path to the water quality result file
            (usually structure_control_actions_3di.nc)
        :param file_modus: modus in which to open the files
        """
        self._netcdf_file_path: str = netcdf_file_path
        super().__init__(h5_file_path, file_modus)

        if swmr:
            self.netcdf_file = H5SwmrFile(netcdf_file_path, file_modus)
        else:
            self.netcdf_file = h5py.File(netcdf_file_path, file_modus)
        self.netcdf_keys = self.netcdf_file.keys()

        self._timeseries_chunk_size = slice(0, DEFAULT_CHUNK_TIMESERIES.stop)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

        # Get substances, sets substances as attributes: cwqa.substance1
        self.substances = []
        for key in self.netcdf_keys:
            regex_match = re.search(r"substance\d+", key)
            if regex_match:
                substance = regex_match.group()
                self.substances.append(substance)
                result_group = self._build_substance_result_group(substance, "")
                self.__setattr__(substance, result_group)
                self._set_substance_attributes_on_result_group(result_group, substance)

        # Get areas, sets areas as attributes: cwqa.area1
        self.areas = []
        self._add_area_groups()

    def get_model_instance_by_field_name(self, field_name):
        """
        :param field_name: name of a models field
        :return: instance of the model the field belongs to
        :raises AttributeError if the model instance cannot be found
        """
        try:
            model_instance = self.__getattribute__(field_name)
            return model_instance
        except AttributeError:
            raise AttributeError(
                f"Model instance with field name {field_name} not found"
            )

    def set_timeseries_chunk_size(self, new_chunk_size: int) -> None:
        """
        overwrite the default chunk size for timeseries queries.
        :param new_chunk_size <int>: new chunk size for
            timeseries queries
        :raises ValueError when the given value is less than 1
        """
        _chunk_size = int(new_chunk_size)
        if _chunk_size < 1:
            raise ValueError("Chunk size must be greater than 0")
        self._timeseries_chunk_size = slice(0, _chunk_size)
        logger.info("New chunk for timeseries size has been set to %d", new_chunk_size)
        self._grid_kwargs.update({"timeseries_chunk_size": self._timeseries_chunk_size})

        # Update timeseries chunk size for all substances
        for substance in self.substances:
            result_group = self.__getattribute__(substance)
            result_group.class_kwargs.update(
                {"timeseries_chunk_size": self._timeseries_chunk_size}
            )

        # Update timeseries chunk size for all substances per area
        for area in self.areas:
            for substance in self.substances:
                result_group = self.__getattribute__(area).__getattribute__(substance)
                result_group.class_kwargs.update(
                    {"timeseries_chunk_size": self._timeseries_chunk_size}
                )

    def _build_substance_result_group(self, substance: str, area: str) -> None:
        return Nodes(
            H5pyResultGroup(self.h5py_file, "nodes", self.netcdf_file),
            **dict(
                self._grid_kwargs,
                **{
                    "mixin": get_nodes_customized_water_quality_results_mixin(
                        substance, self.netcdf_keys, area
                    )
                },
            ),
        )

    def _set_substance_attributes_on_result_group(
        self, result_group: Nodes, substance: str
    ) -> None:
        for key in self.netcdf_keys:
            regex_match = re.search(rf"{substance}_(1|2)D", key)
            if regex_match:
                dataset = regex_match.group()
                attrs_map = [("name", "substance_name"), ("units", "units")]
                for attr, name in attrs_map:
                    value = self.netcdf_file[dataset].attrs.get(name)
                    if isinstance(value, bytes):
                        value = value.decode("utf-8")
                    elif isinstance(value, h5py.Empty):
                        value = ""

                    result_group.__setattr__(attr, value)

    def _add_area_groups(self) -> None:
        for key in self.netcdf_keys:
            regex_match = re.search(r"Mesh\d{1,2}DNode_id_area\d+", key)
            if regex_match:
                # Set area as attribute and create area result admin
                dataset_name = regex_match.group().split("_")[-1]
                area_name = self.netcdf_file[key].attrs.get("area_name", dataset_name)
                if not hasattr(self, dataset_name):
                    self.areas += [dataset_name]
                    self.__setattr__(
                        dataset_name,
                        _CustomizedWaterQualityAreaResultAdmin(
                            self, dataset_name, area_name
                        ),
                    )

    def close(self) -> None:
        super().close()
        self.netcdf_file.close()

    @property
    def nodes(self):
        return None

    @property
    def lines(self):
        return None

    @property
    def breaches(self):
        return None

    @property
    def pumps(self):
        return None

    @property
    def levees(self):
        return None

    @property
    def nodes_embedded(self):
        return None

    @property
    def cross_sections(self):
        return None

    @property
    def cells(self):
        return None


class _CustomizedWaterQualityAreaResultAdmin:
    """Nested class for customized water quality 3Di result files to interact with
    specific areas"""

    def __init__(
        self,
        customized_result_admin: CustomizedWaterQualityResultAdmin,
        dataset_name: str,
        area_name: str,
    ) -> None:
        self.cwqra = customized_result_admin
        self.name = area_name
        self._timeseries_chunk_size = customized_result_admin._timeseries_chunk_size
        self._grid_kwargs = customized_result_admin._grid_kwargs

        for substance in self.cwqra.substances:
            result_group = self.cwqra._build_substance_result_group(
                substance, f"_{dataset_name}"
            )
            self.__setattr__(substance, result_group)
            self.cwqra._set_substance_attributes_on_result_group(
                result_group, substance
            )



"""
Models
++++++

The ``Crosssections`` models represent the cross-sections as used in the
threedicore with shape, width and height information.


"""
from enum import IntEnum, unique

from threedigrid.orm.fields import ArrayField
from threedigrid.orm.models import Model


@unique
class CrossSectionShape(IntEnum):
    CLOSED_RECTANGLE = 0
    RECTANGLE = 1
    CIRCLE = 2
    EGG = 3
    TABULATED_RECTANGLE = 5
    TABULATED_TRAPEZIUM = 6
    TABULATED_YZ = 7
    INVERTED_EGG = 8


class CrossSections(Model):
    code = ArrayField(type=int)
    shape = ArrayField(type=int)
    content_pk = ArrayField(type=int)
    width_1d = ArrayField(type=float)
    offset = ArrayField(type=float)
    count = ArrayField(type=float)
    tables = ArrayField(type=float)
    offset_yz = ArrayField(type=float)
    count_yz = ArrayField(type=float)
    tables_yz = ArrayField(type=float)

    def get_tabulated_cross_section_height_and_width(self, cross_section_pk):
        """
        Returns the heights and widths of a cross-section.

        Args:
            cross_section_pk: (int) the cross-section primary key

        Returns:
            (tuple([heights], [widths])) the heights and widths of the cross-section
        """

        if self.shape[cross_section_pk] == CrossSectionShape.TABULATED_YZ:
            offset = self.offset_yz[cross_section_pk]
            count = self.count_yz[cross_section_pk]
            table = self.tables_yz
        else:
            offset = self.offset[cross_section_pk]
            count = self.count[cross_section_pk]
            table = self.tables

        return (
            table[1, offset : offset + count],
            table[0, offset : offset + count],
        )



from os.path import join

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from threedigrid.admin.gridresultadmin import (
    CustomizedWaterQualityResultAdmin,
    GridH5WaterQualityResultAdmin,
)

from .conftest import test_file_dir


@pytest.fixture
def wqa():
    return GridH5WaterQualityResultAdmin(
        join(test_file_dir, "water_quality/gridadmin.h5"),
        join(test_file_dir, "water_quality/water_quality_results_3di.nc"),
    )


@pytest.fixture
def cwqa():
    return CustomizedWaterQualityResultAdmin(
        join(test_file_dir, "water_quality/gridadmin.h5"),
        join(test_file_dir, "water_quality/customized_water_quality_results_3di.nc"),
    )


def test_concentration_values(wqa):
    assert_array_equal(
        wqa.substance1.concentration[:, 1:], wqa.netcdf_file["substance1_1D"][:]
    )
    assert_array_equal(
        wqa.substance2.concentration[:, 1:], wqa.netcdf_file["substance2_1D"][:]
    )
    assert_array_equal(
        wqa.substance3.concentration[:, 1:], wqa.netcdf_file["substance3_1D"][:]
    )


def test_substance_names(wqa):
    assert wqa.substance1.name == "test1"
    assert wqa.substance2.name == "test2"
    assert wqa.substance3.name == "test3"


def test_units(wqa):
    assert wqa.substance1.units == ""
    assert wqa.substance2.units == "kg/m3"
    assert wqa.substance3.units == "g/ml"


def test_overriden_properties(cwqa: CustomizedWaterQualityResultAdmin):
    assert cwqa.nodes is None
    assert cwqa.lines is None
    assert cwqa.breaches is None
    assert cwqa.pumps is None
    assert cwqa.levees is None
    assert cwqa.nodes_embedded is None
    assert cwqa.cross_sections is None
    assert cwqa.cells is None


def test_substance_names_customized(cwqa: CustomizedWaterQualityResultAdmin):
    assert cwqa.substance1.name == "test1"
    assert cwqa.substance2.name == "test2"
    assert cwqa.substance3.name == "test3"
    assert cwqa.area1.substance1.name == "test1"
    assert cwqa.area1.substance2.name == "test2"
    assert cwqa.area1.substance3.name == "test3"
    assert cwqa.area2.substance1.name == "test1"
    assert cwqa.area2.substance2.name == "test2"
    assert cwqa.area2.substance3.name == "test3"


def test_substance_units_customized(cwqa: CustomizedWaterQualityResultAdmin):
    assert cwqa.substance1.units == ""
    assert cwqa.substance2.units == "kg/m3"
    assert cwqa.substance3.units == "g/ml"
    assert cwqa.area1.substance1.units == ""
    assert cwqa.area1.substance2.units == "kg/m3"
    assert cwqa.area1.substance3.units == "g/ml"
    assert cwqa.area2.substance1.units == ""
    assert cwqa.area2.substance2.units == "kg/m3"
    assert cwqa.area2.substance3.units == "g/ml"


def test_areas_list(cwqa: CustomizedWaterQualityResultAdmin):
    assert cwqa.areas == ["area1", "area2"]
    assert cwqa.area1.name == "test 1"
    assert cwqa.area2.name == "area2"


def test_node_ids(cwqa: CustomizedWaterQualityResultAdmin):
    assert_array_equal(cwqa.substance1.id, np.array([0, 1, 2, 3, 301, 303]))
    assert_array_equal(cwqa.area1.substance2.id, np.array([0, 1, 3, 303]))
    assert_array_equal(cwqa.area2.substance3.id, np.array([0, 2, 301, 303]))


def test_concentration_values_customized(cwqa: CustomizedWaterQualityResultAdmin):
    cwqa.set_timeseries_chunk_size(20)
    assert_array_equal(
        cwqa.substance1.concentration[:, 1:], cwqa.netcdf_file["substance1_1D"][:20]
    )
    assert_array_equal(
        cwqa.substance2.concentration[:, 1:], cwqa.netcdf_file["substance2_1D"][:20]
    )
    assert_array_equal(
        cwqa.substance3.concentration[:, 1:], cwqa.netcdf_file["substance3_1D"][:20]
    )

    cwqa.set_timeseries_chunk_size(30)
    assert_array_equal(
        cwqa.area1.substance1.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance1_1D"][:30], [0, 2, 4], axis=1),
    )
    assert_array_equal(
        cwqa.area1.substance2.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance2_1D"][:30], [0, 2, 4], axis=1),
    )
    assert_array_equal(
        cwqa.area1.substance3.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance3_1D"][:30], [0, 2, 4], axis=1),
    )

    assert_array_equal(
        cwqa.area2.substance1.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance1_1D"][:30], [1, 3, 4], axis=1),
    )
    assert_array_equal(
        cwqa.area2.substance2.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance2_1D"][:30], [1, 3, 4], axis=1),
    )
    assert_array_equal(
        cwqa.area2.substance3.concentration[:, 1:],
        np.take(cwqa.netcdf_file["substance3_1D"][:30], [1, 3, 4], axis=1),
    )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.

from typing import List

from threedigrid.admin.constants import NO_DATA_VALUE
from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin, ResultMixin

BASE_COMPOSITE_FIELDS = {
    "au": ["Mesh2D_au", "Mesh1D_au"],
    "u1": ["Mesh2D_u1", "Mesh1D_u1"],
    "q": ["Mesh2D_q", "Mesh1D_q"],
    "qp": ["Mesh2D_qp"],
    "up1": ["Mesh2D_up1"],
    "_mesh_id": ["Mesh2DLine_id", "Mesh1DLine_id"],  # private
}

BASE_SUBSET_FIELDS = {
    "qp": {"2d_all": "Mesh2D_qp"},
    "up1": {"2d_all": "Mesh2D_up1"},
    "breach_depth": {"1d_all": "Mesh1D_breach_depth"},
    "breach_width": {"1d_all": "Mesh1D_breach_width"},
}


class LinesResultsMixin(ResultMixin):
    class Meta:
        # attributes for the given fields
        field_attrs = ["units", "long_name", "standard_name"]

        composite_fields = BASE_COMPOSITE_FIELDS
        subset_fields = BASE_SUBSET_FIELDS

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


class LinesAggregateResultsMixin(AggregateResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name"]

        base_composition = BASE_COMPOSITE_FIELDS
        base_subset_fields = BASE_SUBSET_FIELDS

        composition_vars = {
            "au": ["min", "max", "avg"],
            "u1": ["min", "max", "avg"],
            "q": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
            "up1": ["min", "max", "avg"],
            "qp": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
        }

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


def get_lines_customized_result_mixin(fields: List[str], area: str):
    def construct_line_customized_base_composite_fields(fields: List[str], area: str):
        """ID is added as a composite field as it is not equal to the grid ids"""
        composite_fields = {}
        for field in fields:
            for key, values in BASE_COMPOSITE_FIELDS.items():
                if field in values:
                    if key not in composite_fields:
                        composite_fields[key] = [field]
                    else:
                        # Sort just to be sure 2D is before 1D
                        composite_fields[key].insert(0, field)
                        composite_fields[key].sort(reverse=True)

        if "_mesh_id" in composite_fields:
            for i, v in enumerate(composite_fields["_mesh_id"]):
                composite_fields["_mesh_id"][i] = f"{v}{area}"

        composite_fields["id"] = []
        if "Mesh2DLine_id" in fields:
            composite_fields["id"] += ["Mesh2DLine_id"]
        if "Mesh1DLine_id" in fields:
            composite_fields["id"] += ["Mesh1DLine_id"]

        composite_fields["kcu"] = []
        if "Mesh2DLine_type" in fields:
            composite_fields["kcu"] += ["Mesh2DLine_type"]
        if "Mesh1DLine_type" in fields:
            composite_fields["kcu"] += ["Mesh1DLine_type"]

        return composite_fields

    def construct_line_customized_base_subset_fields(fields: List[str]):
        subset_fields = {}
        for key, value in BASE_SUBSET_FIELDS.items():
            subset = list(value.keys())[0]
            if value[subset] in fields:
                subset_fields[key] = {subset: value[subset]}

        return subset_fields

    composites = construct_line_customized_base_composite_fields(fields, area)
    subsets = construct_line_customized_base_subset_fields(fields)

    class LinesCustomizedResultsMixin(LinesResultsMixin):
        class Meta:
            field_attrs = ["units", "long_name", "standard_name"]

            composite_fields = composites
            composite_fields_skip_timeseries_filter = ["id", "_mesh_id", "kcu"]
            subset_fields = subsets
            lookup_fields = ("_mesh_id", "id")
            is_customized_mixin = True
            composite_field_insert_values = {"kcu": int(NO_DATA_VALUE)}

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    return LinesCustomizedResultsMixin



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
from typing import List

from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import AggregateResultMixin, ResultMixin

BASE_COMPOSITE_FIELDS = {
    "s1": ["Mesh2D_s1", "Mesh1D_s1"],
    "vol": ["Mesh2D_vol", "Mesh1D_vol"],
    "su": ["Mesh2D_su", "Mesh1D_su"],
    "rain": ["Mesh2D_rain", "Mesh1D_rain"],
    "q_lat": ["Mesh2D_q_lat", "Mesh1D_q_lat"],
    "_mesh_id": ["Mesh2DNode_id", "Mesh1DNode_id"],
}

BASE_SUBSET_FIELDS = {
    "infiltration_rate_simple": {"2d_all": "Mesh2D_infiltration_rate_simple"},
    "ucx": {"2d_all": "Mesh2D_ucx"},
    "ucy": {"2d_all": "Mesh2D_ucy"},
    "leak": {"2d_all": "Mesh2D_leak"},
    "intercepted_volume": {"2d_all": "Mesh2D_intercepted_volume"},
    "q_sss": {"2d_all": "Mesh2D_q_sss"},
}


class NodesResultsMixin(ResultMixin):
    class Meta:
        # attributes for the given fields
        field_attrs = ["units", "long_name", "standard_name"]

        # values of *_COMPOSITE_FIELDS are the variables names as known in
        # the result netCDF file. They are split into 1D and 2D subsets.
        # As threedigrid has its own subsection ecosystem they are merged
        # into a single field (e.g. the keys of *_COMPOSITE_FIELDS).

        # N.B. # fields starting with '_' are private and will not be added to
        # fields property
        composite_fields = BASE_COMPOSITE_FIELDS
        subset_fields = BASE_SUBSET_FIELDS

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a node with netcdf results.

        Variables stored in the netcdf and related to nodes are dynamically
        added as attributes as TimeSeriesCompositeArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


class NodesAggregateResultsMixin(AggregateResultMixin):
    class Meta(metaclass=ModelMeta):
        base_composition = BASE_COMPOSITE_FIELDS
        base_subset_fields = BASE_SUBSET_FIELDS

        # attributes for the given fields
        field_attrs = ["units", "long_name"]

        # extra vars that will be combined with the
        # composite fields, e.g.
        # s1 --> s1_min [Mesh2D_s1_min + Mesh1D_s1_min]
        #    --> s1_max  [Mesh2D_s1_max + Mesh1D_s1_max]
        composition_vars = {
            "s1": ["min", "max", "avg"],
            "vol": ["min", "max", "avg", "sum", "current"],
            "su": ["min", "max", "avg"],
            "rain": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
            "q_lat": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
            "infiltration_rate_simple": [
                "min",
                "max",
                "avg",
                "cum",
                "cum_positive",
                "cum_negative",
            ],
            "ucx": ["min", "max", "avg"],
            "ucy": ["min", "max", "avg"],
            "leak": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
            "intercepted_volume": ["min", "max", "avg", "cum", "current"],
            "q_sss": ["min", "max", "avg", "cum", "cum_positive", "cum_negative"],
        }

        lookup_fields = ("id", "_mesh_id")

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


class NodeSubstanceResultMixinBase(ResultMixin):
    def __init__(self, **kwargs):
        """Instantiate a node with netcdf results.

        Variables stored in the netcdf and related to nodes are dynamically
        added as attributes as TimeSeriesCompositeArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


def get_substance_result_mixin(substance_name: str):
    class NodeSubstanceResultMixin(NodeSubstanceResultMixinBase):
        class Meta:
            # attributes for the given fields
            field_attrs = ["long_name"]

            # values of *_COMPOSITE_FIELDS are the variables names as known in
            # the result netCDF file. They are split into 1D and 2D subsets.
            # As threedigrid has its own subsection ecosystem they are merged
            # into a single field (e.g. the keys of *_COMPOSITE_FIELDS).

            # N.B. # fields starting with '_' are private and will not be added to
            # fields property
            composite_fields = {
                "concentration": [f"{substance_name}_2D", f"{substance_name}_1D"],
                "_mesh_id": ["Mesh2DNode_id", "Mesh1DNode_id"],
            }
            subset_fields = {}

            lookup_fields = ("id", "_mesh_id")

    return NodeSubstanceResultMixin


def get_nodes_customized_results_mixin(fields: List[str], area: str):
    def construct_node_customized_base_composite_fields(fields: List[str], area: str):
        """ID is added as a composite field as it is not equal to the grid ids"""
        composite_fields = {}
        for field in fields:
            for key, values in BASE_COMPOSITE_FIELDS.items():
                if field in values:
                    if key not in composite_fields:
                        composite_fields[key] = [field]
                    else:
                        # Sort just to be sure 2D is before 1D
                        composite_fields[key].insert(0, field)
                        composite_fields[key].sort(reverse=True)

        if "_mesh_id" in composite_fields:
            for i, v in enumerate(composite_fields["_mesh_id"]):
                composite_fields["_mesh_id"][i] = f"{v}{area}"

        composite_fields["id"] = []
        if "Mesh2DNode_id" in fields:
            composite_fields["id"] += ["Mesh2DNode_id"]
        if "Mesh1DNode_id" in fields:
            composite_fields["id"] += ["Mesh1DNode_id"]

        composite_fields["node_type"] = []
        if "Mesh2DNode_type" in fields:
            composite_fields["node_type"] += ["Mesh2DNode_type"]
        if "Mesh1DNode_type" in fields:
            composite_fields["node_type"] += ["Mesh1DNode_type"]

        return composite_fields

    def construct_node_customized_base_subset_fields(fields: List[str]):
        subset_fields = {}
        for key, value in BASE_SUBSET_FIELDS.items():
            if value["2d_all"] in fields:
                subset_fields[key] = {"2d_all": value["2d_all"]}

        return subset_fields

    composites = construct_node_customized_base_composite_fields(fields, area)
    subsets = construct_node_customized_base_subset_fields(fields)

    class NodesCustomizedResultsMixin(ResultMixin):
        class Meta:
            field_attrs = ["units", "long_name", "standard_name"]

            composite_fields = composites
            composite_fields_skip_timeseries_filter = ["id", "_mesh_id", "node_type"]
            subset_fields = subsets
            lookup_fields = ("_mesh_id", "id")
            is_customized_mixin = True

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    return NodesCustomizedResultsMixin


def get_nodes_customized_water_quality_results_mixin(
    substance_name: str, fields: List[str], area: str
):
    composites = {}
    composites["id"] = []
    if "Mesh2DNode_id" in fields:
        composites["id"] += ["Mesh2DNode_id"]
    if "Mesh1DNode_id" in fields:
        composites["id"] += ["Mesh1DNode_id"]

    composites["_mesh_id"] = []
    if f"Mesh2DNode_id{area}" in fields:
        composites["_mesh_id"] += [f"Mesh2DNode_id{area}"]
    if f"Mesh1DNode_id{area}" in fields:
        composites["_mesh_id"] += [f"Mesh1DNode_id{area}"]

    composites["node_type"] = []
    if "Mesh2DNode_type" in fields:
        composites["node_type"] += ["Mesh2DNode_type"]
    if "Mesh1DNode_type" in fields:
        composites["node_type"] += ["Mesh1DNode_type"]

    composites["concentration"] = []
    if f"{substance_name}_2D" in fields:
        composites["concentration"] += [f"{substance_name}_2D"]
    if f"{substance_name}_1D" in fields:
        composites["concentration"] += [f"{substance_name}_1D"]

    class NodesCustomizedWaterQualityResultsMixin(ResultMixin):
        class Meta:
            field_attrs = ["units", "long_name", "standard_name"]

            composite_fields = composites
            composite_fields_skip_timeseries_filter = ["id", "_mesh_id", "node_type"]

            subset_fields = {}
            lookup_fields = ("_mesh_id", "id")
            is_customized_mixin = True

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    return NodesCustomizedWaterQualityResultsMixin



from unittest import mock

from threedigrid.admin.constants import TYPE_CODE_MAP
from threedigrid.admin.idmapper import IdMapper

from .conftest import NODE_LENGTH, simple_id_map


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_prepare_mapper(mocked_id_map, h5py_file, threedi_datasource):
    mocked_id_map.return_value = simple_id_map()
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    # now a mapping should exist on the h5py file
    mapping = h5py_file["mappings"]["id_map"]
    assert mapping
    assert mapping.dtype.names == ("obj_code", "pk", "seq_id")
    assert mapping[:].size == len(TYPE_CODE_MAP.values()) * NODE_LENGTH


@mock.patch("threedigrid.admin.idmapper.get_id_map")
def test_init_idMapper(mocked_id_map, h5py_file, threedi_datasource):
    mocked_id_map.return_value = simple_id_map()
    IdMapper.prepare_mapper(h5py_file, threedi_datasource)
    mapper = IdMapper(h5py_file["mappings"]["id_map"])
    assert mapper.get_by_code(TYPE_CODE_MAP["v2_channel"]).size == NODE_LENGTH
    assert mapper.get_by_name("v2_channel").size == NODE_LENGTH



from os.path import join

import numpy as np
import pytest
from numpy.testing import assert_array_equal

from threedigrid.admin.gridresultadmin import CustomizedResultAdmin
from threedigrid.numpy_utils import create_np_lookup_index_for

from .conftest import test_file_dir


@pytest.fixture
def grc():
    grc = CustomizedResultAdmin(
        join(test_file_dir, "bergen_with_boundaries/gridadmin.h5"),
        join(test_file_dir, "bergen_with_boundaries/customized_results_3di.nc"),
    )
    yield grc
    grc.close()


@pytest.fixture
def grc_breach():
    grc = CustomizedResultAdmin(
        join(test_file_dir, "breach_growth/gridadmin.h5"),
        join(test_file_dir, "breach_growth/customized_results_3di.nc"),
    )
    yield grc
    grc.close()


def test_overriden_grc_properties(grc: CustomizedResultAdmin):
    assert grc.levees is None
    assert grc.nodes_embedded is None
    assert grc.cross_sections is None
    assert grc.cells is None


def test_grc_areas(grc: CustomizedResultAdmin):
    assert grc.areas == ["area1", "area2"]
    assert grc.area1.name == "test 1"
    assert grc.area2.name == "area2"


def test_grc_root_node_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.nodes.id,
        [0, 2077, 2078, 2091, 2092, 7730, 7731, 7744, 7745, 12064, 12311, 12545],
    )
    assert_array_equal(
        grc.nodes.subset("2D_ALL").id,
        [2077, 2078, 2091, 2092, 7730, 7731, 7744, 7745],
    )
    assert_array_equal(
        grc.nodes.subset("1D_ALL").id,
        [12064, 12311, 12545],
    )


@pytest.mark.parametrize("field", ["s1", "vol", "su", "rain", "q_lat"])
def test_grc_root_node_results_composites(grc: CustomizedResultAdmin, field):
    """Check composite fields for base nodes attribute"""
    assert_array_equal(
        getattr(grc.nodes, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10],
                grc.netcdf_file[f"Mesh1D_{field}"][:10],
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(grc.nodes.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10],
    )
    assert_array_equal(
        getattr(grc.nodes.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10],
    )


@pytest.mark.parametrize("field", ["ucx", "ucy", "q_sss"])
def test_grc_root_node_results_subsets(grc: CustomizedResultAdmin, field):
    """Check subset fields for base nodes attribute"""
    assert_array_equal(
        getattr(grc.nodes, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10],
                np.zeros((10, 3), dtype=np.float32),
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(grc.nodes.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10],
    )


def test_grc_area1_node_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area1.nodes.id,
        [0, 2078, 2092, 7731, 7745, 12064, 12311, 12545],
    )
    assert_array_equal(
        grc.area1.nodes.subset("2D_ALL").id,
        [2078, 2092, 7731, 7745],
    )
    assert_array_equal(
        grc.area1.nodes.subset("1D_ALL").id,
        [12064, 12311, 12545],
    )


def test_grc_area2_node_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area2.nodes.id,
        [0, 2077, 2091, 7730, 7744],
    )
    assert_array_equal(
        grc.area2.nodes.subset("2D_ALL").id,
        [2077, 2091, 7730, 7744],
    )
    assert_array_equal(
        grc.area2.nodes.subset("1D_ALL").id,
        [],
    )


@pytest.mark.parametrize("field", ["s1", "vol", "su", "rain", "q_lat"])
@pytest.mark.parametrize("area", ["area1", "area2"])
def test_grc_node_areas_results_composites(grc: CustomizedResultAdmin, field, area):
    """Check composite fields for areas nodes attribute"""
    subset_index_2d = create_np_lookup_index_for(
        getattr(grc, area).nodes.subset("2D_ALL").id,
        grc.nodes.subset("2D_ALL").id,
    )
    subset_index_1d = create_np_lookup_index_for(
        getattr(grc, area).nodes.subset("1D_ALL").id, grc.nodes.subset("1D_ALL").id
    )
    assert_array_equal(
        getattr(getattr(grc, area).nodes, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
                grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(getattr(grc, area).nodes.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
    )
    assert_array_equal(
        getattr(getattr(grc, area).nodes.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
    )


@pytest.mark.parametrize("field", ["ucx", "ucy", "q_sss"])
@pytest.mark.parametrize("area", ["area1", "area2"])
def test_grc_node_area_results_subsets(grc: CustomizedResultAdmin, field, area):
    """Check subset fields for area nodes attribute"""
    subset_index_2d = create_np_lookup_index_for(
        getattr(grc, area).nodes.subset("2D_ALL").id,
        grc.nodes.subset("2D_ALL").id,
    )
    one_d_size = getattr(grc, area).nodes.subset("1D_ALL").count

    assert_array_equal(
        getattr(getattr(grc, area).nodes, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
                np.zeros((10, one_d_size), dtype=np.float32),
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(getattr(grc, area).nodes.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
    )
    assert_array_equal(
        getattr(getattr(grc, area).nodes.subset("1D_ALL"), field),
        [],
    )


def test_grc_root_line_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.lines.id,
        [0, 5418, 11150, 22635, 28367, 29085, 29629],
    )
    assert_array_equal(
        grc.lines.subset("2D_ALL").id,
        [5418, 11150, 22635, 28367],
    )
    assert_array_equal(
        grc.lines.subset("1D_ALL").id,
        [29085, 29629],
    )


def test_grc_area1_line_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area1.lines.id,
        [0, 28367, 29085, 29629],
    )
    assert_array_equal(
        grc.area1.lines.subset("2D_ALL").id,
        [28367],
    )
    assert_array_equal(
        grc.area1.lines.subset("1D_ALL").id,
        [29085, 29629],
    )


def test_grc_area2_line_ids(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area2.lines.id,
        [0, 5418, 11150, 22635],
    )
    assert_array_equal(
        grc.area2.lines.subset("2D_ALL").id,
        [5418, 11150, 22635],
    )
    assert_array_equal(
        grc.area2.lines.subset("1D_ALL").id,
        [],
    )


@pytest.mark.parametrize("field", ["u1", "q", "au"])
def test_grc_root_line_results_composites(grc: CustomizedResultAdmin, field):
    """Check composite fields for base lines attribute"""
    assert_array_equal(
        getattr(grc.lines, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10],
                grc.netcdf_file[f"Mesh1D_{field}"][:10],
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(grc.lines.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10],
    )
    assert_array_equal(
        getattr(grc.lines.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10],
    )


@pytest.mark.parametrize("field", ["breach_depth", "breach_width"])
def test_grc_root_line_results_subsets(grc: CustomizedResultAdmin, field):
    """Check subset fields for base lines attribute"""
    assert_array_equal(
        getattr(grc.lines, field)[:, 1:],
        np.concatenate(
            (
                np.zeros((10, 4), dtype=np.float32),
                grc.netcdf_file[f"Mesh1D_{field}"][:10],
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(grc.lines.subset("2D_ALL"), field),
        np.zeros((10, 4), dtype=np.float32),
    )
    assert_array_equal(
        getattr(grc.lines.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10],
    )


@pytest.mark.parametrize("field", ["u1", "q", "au"])
@pytest.mark.parametrize("area", ["area1", "area2"])
def test_grc_area_line_results_composites(grc: CustomizedResultAdmin, field, area):
    """Check composite fields for area lines attribute"""
    subset_index_2d = create_np_lookup_index_for(
        getattr(grc, area).lines.subset("2D_ALL").id,
        grc.lines.subset("2D_ALL").id,
    )
    subset_index_1d = create_np_lookup_index_for(
        getattr(grc, area).lines.subset("1D_ALL").id, grc.lines.subset("1D_ALL").id
    )
    assert_array_equal(
        getattr(getattr(grc, area).lines, field)[:, 1:],
        np.concatenate(
            (
                grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
                grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
            ),
            axis=1,
        ),
    )
    assert_array_equal(
        getattr(getattr(grc, area).lines.subset("2D_ALL"), field),
        grc.netcdf_file[f"Mesh2D_{field}"][:10, subset_index_2d],
    )
    assert_array_equal(
        getattr(getattr(grc, area).lines.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
    )


@pytest.mark.parametrize("field", ["breach_depth", "breach_width"])
def test_grc_line_area1_1d_results(grc, field):
    """Check subset fields for base lines attribute"""
    subset_index_2d = create_np_lookup_index_for(
        grc.area1.lines.subset("2D_ALL").id,
        grc.lines.subset("2D_ALL").id,
    )
    subset_index_1d = create_np_lookup_index_for(
        grc.area1.lines.subset("1D_ALL").id, grc.lines.subset("1D_ALL").id
    )

    assert_array_equal(
        getattr(grc.area1.lines, field)[:, 1:],
        np.concatenate(
            (
                np.zeros((10, subset_index_2d.size), dtype=np.float32),
                grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
            ),
            axis=1,
        ),
    )

    assert_array_equal(
        getattr(grc.area1.lines.subset("2D_ALL"), field),
        np.zeros((10, subset_index_2d.size), dtype=np.float32),
    )

    assert_array_equal(
        getattr(grc.area1.lines.subset("1D_ALL"), field),
        grc.netcdf_file[f"Mesh1D_{field}"][:10, subset_index_1d],
    ),


@pytest.mark.parametrize("field", ["breach_depth", "breach_width"])
def test_grc_line_area2_1d_results(grc, field):
    """No area 2 1D"""
    assert_array_equal(getattr(grc.area2.lines, field)[:], [])

    assert_array_equal(
        getattr(grc.area2.lines.subset("2D_ALL"), field),
        [],
    )

    assert_array_equal(getattr(grc.area2.lines.subset("1D_ALL"), field), []),


def test_pumps_root(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.pumps.id,
        [0, 1, 2],
    )
    assert_array_equal(
        grc.pumps.q_pump[:, 1:],
        grc.netcdf_file["Mesh1D_q_pump"][:10],
    )


def test_pumps_area1(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area1.pumps.id,
        [0, 1],
    )
    assert_array_equal(
        grc.area1.pumps.q_pump[:, 1:].flatten(),
        grc.netcdf_file["Mesh1D_q_pump"][:10, 0],
    )


def test_pumps_area2(grc: CustomizedResultAdmin):
    assert_array_equal(
        grc.area2.pumps.id,
        [0, 2],
    )
    assert_array_equal(
        grc.area2.pumps.q_pump[:, 1:].flatten(),
        grc.netcdf_file["Mesh1D_q_pump"][:10, 1],
    )


def test_breaches_root(grc_breach: CustomizedResultAdmin):
    grc_breach.set_timeseries_chunk_size(50)
    assert_array_equal(
        grc_breach.breaches.id,
        [0, 764, 765],
    )
    assert_array_equal(
        grc_breach.breaches.timestamps,
        grc_breach.netcdf_file["time"][:],
    )
    assert_array_equal(
        grc_breach.breaches.breach_depth[:, 1:],
        grc_breach.netcdf_file["Mesh1D_breach_depth"][:],
    )
    assert_array_equal(
        grc_breach.breaches.breach_width[:, 1:],
        grc_breach.netcdf_file["Mesh1D_breach_width"][:],
    )


def test_breaches_area1(grc_breach: CustomizedResultAdmin):
    grc_breach.set_timeseries_chunk_size(50)
    assert_array_equal(
        grc_breach.area1.breaches.id[1:],
        [764],
    )
    assert_array_equal(
        grc_breach.area1.breaches.breach_depth[:, 1:].flatten(),
        grc_breach.netcdf_file["Mesh1D_breach_depth"][:, 0],
    )
    assert_array_equal(
        grc_breach.area1.breaches.breach_width[:, 1:].flatten(),
        grc_breach.netcdf_file["Mesh1D_breach_width"][:, 0],
    )


def test_breaches_area2(grc_breach: CustomizedResultAdmin):
    grc_breach.set_timeseries_chunk_size(50)
    assert_array_equal(
        grc_breach.area2.breaches.id[1:],
        [765],
    )
    assert_array_equal(
        grc_breach.area2.breaches.breach_depth[:, 1:].flatten(),
        grc_breach.netcdf_file["Mesh1D_breach_depth"][:, 1],
    )
    assert_array_equal(
        grc_breach.area2.breaches.breach_width[:, 1:].flatten(),
        grc_breach.netcdf_file["Mesh1D_breach_width"][:, 1],
    )



# (c) Nelen & Schuurmans.  GPL licensed, see LICENSE.rst.
from typing import List

from threedigrid.admin.constants import NO_DATA_VALUE
from threedigrid.orm.base.options import ModelMeta
from threedigrid.orm.base.timeseries_mixin import ResultMixin


class BreachesResultsMixin(ResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name", "standard_name"]

        # customize field names
        composite_fields = {
            "breach_depth": ["Mesh1D_breach_depth"],
            "breach_width": ["Mesh1D_breach_width"],
        }

    def __init__(self, **kwargs):
        """Instantiate a breach with netcdf results.

        Variables stored in the netcdf and related to breaches are dynamically
        added as attributes as TimeSeriesArrayField.

        :param kwargs:
        """
        super().__init__(**kwargs)


class BreachesAggregateResultsMixin(ResultMixin):
    class Meta(metaclass=ModelMeta):
        field_attrs = ["units", "long_name"]

        base_composition = {
            "breach_depth": ["Mesh1D_breach_depth"],
            "breach_width": ["Mesh1D_breach_width"],
        }
        composition_vars = {
            "breach_depth": ["avg", "min", "max"],
            "breach_width": ["avg", "min", "max"],
        }

    def __init__(self, **kwargs):
        """Instantiate a line with netcdf results.

        Variables stored in the netcdf and related to lines are dynamically
        added as attributes as TimeSeriesArrayField.

        :param netcdf_keys: list of netcdf variables
        :param kwargs:
        """
        super().__init__(**kwargs)


def get_breaches_customized_result_mixin(fields: List[str], area: str):
    if f"Mesh1DLine_id{area}" not in fields:
        return

    composites = {}
    composites["id"] = ["Mesh1DLine_id"]
    composites["_mesh_id"] = [f"Mesh1DLine_id{area}"]
    if "Mesh1D_breach_depth" in fields:
        composites["breach_depth"] = ["Mesh1D_breach_depth"]
    if "Mesh1D_breach_width" in fields:
        composites["breach_width"] = ["Mesh1D_breach_width"]

    class BreachesResultsCustomizedResultMixin(BreachesResultsMixin):
        class Meta:
            field_attrs = ["units", "long_name", "standard_name"]
            composite_fields = composites
            lookup_fields = ("_mesh_id", "levl")
            is_customized_mixin = True
            composite_field_insert_values = {"_mesh_id": int(NO_DATA_VALUE)}

        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    return BreachesResultsCustomizedResultMixin
